/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1993 Electrotechnical Laboratry (ETL)

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	onewWnn (Wnn kanakan interface)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	931104	extracted from onewlib.c
///////////////////////////////////////////////////////////////////////*/
int KanakanExists;
#define STATIC static

#define VERSION	Onew_version()

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include "onew.h"
#include "WnnConf.h"
#include "Wnn_euc.h"

char *strtok();

/*######################################################################*
 *	Wnn STARTUP							*
 *######################################################################*/
typedef struct bunjoho BUN_JOHO;
BUN_JOHO *jd_open_in();
static BUN_JOHO	*BunJoho;
static int	NoOfBunsetu;

#define S_KANALEN	2000	/* jd_open() */
#define S_KLISTLEN	500	/* jd_open() */
#define S_KANJILEN	2000	/* jd_open(),jd_recove(),jd_tanconv() */
#define MAX_ALTS	500	/* jd_next() */
#define ALT_KANJILEN	5000	/* jd_next() */

static BUN_JOHO*
kanakan_init(){
	int omask,diclisti,dicformi,pri;
	Pathname hostname;
	char *jshost, *user, *dl, *dp, *dicf;
	Pathname diclist,dname,dicname,dicform,dicdir,frqdir,dicpath,frqpath;

	if(BunJoho) return BunJoho;

	jshost = getenv("JSERVER");
	if( jshost == 0 ){
		gethostname(hostname,sizeof(hostname));
		jshost = hostname;
	}
	putmsg(1,"%s ** JSERVER(%s)%s",VERSION,jshost,ONEW_msg_connecting);

	user = getenv("USER");
	BunJoho = jd_open_in(S_KANALEN,S_KLISTLEN,S_KANJILEN,jshost,user,0);
	if(BunJoho == 0){
		putmsg(1,"%s ** JSERVER(%s)%s",VERSION,jshost,
			ONEW_msg_connectfail);
		ONEW_BEEP(2);
		return 0;
	}
	for( diclisti = 0; dl = ONEW_DICLIST[diclisti]; diclisti++ ){
		Onew_env_substitute(dl,diclist);
		if( index(diclist,'$') == 0 )
			break;
	}
	/*omask = sigsetmask(-1);*/
	omask = sigblock(sigmask(SIGALRM));
	for( dp=strtok(diclist," \t\n"); dp; dp=strtok(0," \t\n") ){
		pri = 1;
		sscanf(dp,"%[^=]=%d",dname,&pri);
		Onew_env_substitute(dname,dicname);
		if( index(dicname,'/') ){
			if( jd_dicadd(dicname,"",pri,0) != -1 )
				goto NEXT;
			strcpy(dicpath,dicname);
			strcpy(frqpath,"");
		}else
		for(dicformi=0; dicf = ONEW_DICDIRS[dicformi]; dicformi++){
			Onew_env_substitute(dicf,dicform);
			*dicdir = *frqdir = 0;
			sscanf(dicform,"%[^;];%s",dicdir,frqdir);
			if( index(dicdir,'%') == 0 )
				continue;
			sprintf(dicpath,dicdir,dicname,"","");
			sprintf(frqpath,frqdir,dicname,"","");
			if( *dicpath == 0 )
				continue;
			if( jd_dicadd(dicpath,frqpath,pri,0) != -1 )
				goto NEXT;
			putmsg(1,"%s;%s%s",dicpath,frqpath,ONEW_msg_dicfail);
		}
		putmsg(1,"%s;%s%s",dicpath,frqpath,ONEW_msg_dicfail);
		ONEW_BEEP(2);
	NEXT:;
	}
	sigsetmask(omask);
	return BunJoho;
}
static ONEWkanakan_ready(){
	return kanakan_init() != 0;
}
static ONEWkanakan(com,str)
	Uchar *str;
{	int comch;
	BUN_JOHO *BJ;
	char yomibuf[S_KANALEN];

	BJ = kanakan_init();
	if( BJ == 0 ) return 0;
	clr_keyinQ();

	Onew_setupYomistr(com,str,yomibuf);
	ToServerString(yomibuf,BJ->kana_buf);
	comch = kana_cands(BJ);
	return comch;
}


static int henkan_sareta;
static ONEWkanakan_freqsave(){
	if( !henkan_sareta )
		return 0;
	henkan_sareta = 0;
	if(BunJoho)
	if(jd_freqsv() != 0){
		putmsg(1,ONEW_msg_cantsavehind);
		ONEW_BEEP(2);
		return -1;
	}
	return 0;
}

static kanakan_end(){
	if(BunJoho){
		BunJoho = 0;
		if(jd_close()  != 0){
			putmsg(1,">> wnn lib close err");
			ONEW_BEEP(2);
		}
	}
}

STATIC char*
BunsetsuString(BJ,i,out)
	BUN_JOHO *BJ;
	char *out;
{	struct kouho_entry *kep;
	Uchar *sp;

	kep = &BJ->klist[i];
	sp = ToLocalString(out,kep->k_data,wchar_strlen(kep->k_data));
	ToLocalString(sp,&BJ->kana_buf[kep->s_ichi+kep->jl],kep->fl);
	return out;
}
static wchar*
ONEWkanakan_kouho(alts,i,out)
	struct jikouhojoho *alts;
	wchar *out;
{	struct jikouho_entry *jkep;

	jkep = alts->jlist;
	ToLocalString(out,jkep[i].k_data,wchar_strlen(jkep[i].k_data));
	return out;
}

STATIC wchar*
YomiString(BJ,i,out)
	BUN_JOHO *BJ;
	wchar *out;
{	int len;

	*out = 0;
	len = BJ->klist[i].jl + BJ->klist[i].fl;
	ToLocalString(out,&BJ->kana_buf[BJ->klist[i].s_ichi],len);
}
static int
ONEWbunsetu_next(index,inc){
	int i;
	wchar yomi[1000];
	Uchar *cp;

	i = index;
	for( ;; ){
		i += inc;
		if( i < 0 )		{ i = 0; break; }
		if( i >= NoOfBunsetu )	{ i = NoOfBunsetu - 1; break; }
		if( i == index ) break; 
		YomiString(BunJoho,i,yomi);

		for(cp = (Uchar*)yomi; *cp; cp++)
			if( *cp & 0x80 )
				goto OUT;
	}
OUT:	return i;
}

/*######################################################################*
 *	DISPLAY kanji_buff WITH WORD CURSOR				*
 *######################################################################*/
STATIC OW_kanji_buf_disp(BJ,cur_bi,so)
	BUN_JOHO *BJ;
{	int i,col;
	Uchar left[S_KANALEN],current[S_KANALEN],right[S_KANALEN];

	left[0] = 0;
	for(i = 0; i < cur_bi && i < NoOfBunsetu; i++)
		BunsetsuString(BJ,i,&left[strlen(left)]);

	current[0] = 0;
	if( i == cur_bi )
		BunsetsuString(BJ,i,current);

	right[0] = 0;
	for(i++; i < NoOfBunsetu; i++)
		BunsetsuString(BJ,i,&right[strlen(right)]);

	Onew_dispKanakanBuf(so,left,current,right);
}


/*######################################################################*
 *	KANJI_BUFFER EDITOR						*
 *######################################################################*/
static OW_CurrentSerial;
static OW_CurrentDic_no;

STATIC kana_cands(BJ)
	BUN_JOHO *BJ;
{	struct kouho_entry *kep;
	int bunsetux,altx,candn;
	struct jikouhojoho altcand;
	struct jikouho_entry altcands[MAX_ALTS],*jkep;
	wchar altklist[ALT_KANJILEN];
	char comch;

	NoOfBunsetu = jd_begin(BJ->kanji_buf,BJ->kanji_buf_size);
	for(bunsetux = ONEWbunsetu_next(-1,1); bunsetux < NoOfBunsetu; ){
		if( NoOfBunsetu <= 0 ){
			putmsg(1,"ERROR: jserver dead ? (BunsetuNo==0)");
			ONEW_BEEP(2);
			BunJoho = 0;
			return 0x7F;
		}
		if(NoOfBunsetu <= bunsetux)
			bunsetux = NoOfBunsetu-1;
		if(bunsetux < 0)
			bunsetux = NoOfBunsetu-1;

		kep = &BJ->klist[bunsetux];
		OW_kanji_buf_disp(BJ,bunsetux,1);

		altcand.jlist = altcands;
		altcand.jlist_size = MAX_ALTS;
		altcand.kanji_buf = altklist;
		altcand.kanji_buf_size = ALT_KANJILEN;
		candn = jd_next(bunsetux,&altcand);

		altx = 0;
		for(;;){
			OW_CurrentDic_no = kep->jishono;
			OW_CurrentSerial = kep->serial;
			onew_put_kanakanmode();
			comch = OnewGetchar();
			{
			    int rcode;
			    rcode = onew_kanakancom(BJ,comch,&bunsetux,&altx,
					candn,kep,&altcand);
			    switch( rcode ){
				case ONEW_CONT:	break;
				case ONEW_LOOP:	goto Cont;
				case ONEW_EXIT:	goto Exit;
			    }
			}
			if(candn <= altx) altx = 0;
			if(altx < 0) altx = candn - 1;

			jkep = &altcand.jlist[altx];
			kep->jishono = jkep->jishono;
			kep->serial  = jkep->serial;

			if( strcmp(kep->k_data,jkep->k_data) ){
				kep->fl = (kep->jl + kep->fl) - jkep->jl;
				kep->jl = jkep->jl; 
				wchar_strcpy(kep->k_data,jkep->k_data);

				if( (bunsetux+1) < NoOfBunsetu ){ 
				    int inbun;

				    inbun = jd_reconv(bunsetux+1,
				     &kep->k_data[wchar_strlen(kep->k_data)+1],
						S_KANJILEN);
				    NoOfBunsetu = bunsetux+1 + inbun;
				}
				OW_kanji_buf_disp(BJ,bunsetux,1);
			}
		} Cont:;
	} Exit:

	if( jd_end() != 0 ){
		putmsg(1,ONEW_msg_cantsavehind);;
		ONEW_BEEP(1);
	}
	henkan_sareta++;
	OW_kanji_buf_disp(BJ,1000,0);
	return comch;
}

static ONEWbunsetu_expand(BJ,kep,i)
	BUN_JOHO *BJ;
	struct kouho_entry *kep;
{	int inbun,wleng;
	wchar *kbp;

	kbp = BJ->kana_buf;
	wleng = kep->jl + kep->fl;
	if( kbp[kep->s_ichi+wleng] ){
		inbun = jd_tanconv(i,wleng+1,kep->k_data,S_KANJILEN);
		if(0 < inbun)
			NoOfBunsetu = i + inbun;
	}
}
static ONEWbunsetu_shrink(BJ,kep,i)
	BUN_JOHO *BJ;
	struct kouho_entry *kep;
{	int inbun,wleng;

	wleng = kep->jl + kep->fl;
	if(1 < wleng ){
		inbun = jd_tanconv(i,wleng-1,kep->k_data,S_KANJILEN);
		if(0 < inbun)
			NoOfBunsetu = i + inbun;
	}
}

static ONEWbunsetu_delete(BJ,dbunsetu)
	BUN_JOHO *BJ;
{	struct kouho_entry *kep;
	wchar *kbp;
	int i,wleng;

	kbp = BJ->kana_buf;
	kep = &BJ->klist[dbunsetu];
	wleng = kep->jl + kep->fl;

	for(i = kep->s_ichi; kbp[i + wleng]; i++)
		kbp[i] = kbp[i + wleng];
	kbp[i] = 0;

	for(i = dbunsetu; i < NoOfBunsetu; i++){
		kep = &BJ->klist[i];
		kep[0] = kep[1];
		kep->s_ichi -= wleng;
	}
	NoOfBunsetu--;
}


static add_alts(BJ,alts)
	BUN_JOHO *BJ;
	struct jikouhojoho *alts;
{	struct kouho_entry *kep;
	wchar *kbp;
	FILE *fp;
	char command[1024],exp[1024],result[2048];

	kbp = BJ->kana_buf;
	kep = &BJ->klist[0];
	strncpy(exp,kbp,256);
	Onew_rmpadds(exp);
	sprintf(command,"sh -c \"echo '%s' | bc\"",exp);
	if( fp = popen(command,"r") ){
		fgets(result,sizeof(result),fp);
		pclose(fp);
		strcpy(alts->kanji_buf,result);
		kep->jl = strlen(exp);
		kep->fl = 0;
	}else	sprintf(alts->kanji_buf,"Fail popen(%s)",command);
}
static ONEWexternal_henkan(BJ,alts){
	add_alts(BJ,alts);
}


static wchar_strcpy(dst,src)
	wchar *dst,*src;
{	register wchar *d,*s;

	d = dst;
	s = src;
	while( *d++ = *s++ );
}
static wchar_strlen(wstr) wchar *wstr; {
	int len;
	wchar *wcp;

	len = 0;
	for(wcp = wstr; *wcp; wcp++)
		len++;
	return len;
}

static get_dicno(name,dicno)
	char *name;
{	int ndic,dici,udic;
	struct dicinfo *di;
	struct dicinfo dicinfo[64];
	char dicnames[4096];
	char *dname;
	
	ndic = jd_dicinfo(dicinfo,64,dicnames,4096);
	for( dici = 0; dici < ndic; dici++ ){
		di = &dicinfo[dici];
		if( dname = (char*)rindex(di->file_name,'/') )
			dname++;
		else	dname = di->file_name;

		if( strcmp(name,"USERDIC1") == 0 ){
			if( jd_udp(di->dic_no) & 1 ){
				strcpy(name,dname);
				return di->dic_no;
			}
		}else
		if( *name != 0 ){
			if( strcmp(name,dname) == 0 )
				return di->dic_no;
		}else{
			if( dicno == di->dic_no ){
				strcpy(name,dname);
				return di->dic_no;
			}
		}
	}
	return -1;
}
static change_udic(name)
{	int dicno;

	if( 0 <= (dicno = get_dicno(name,0)) ){
		if( jd_udchg(dicno) == 0 )
			return dicno;
		putmsg(1,"UD (%s) CHANGE FAILED",name);
	}else	putmsg(1,"UD (%s) NOT FOUND",name);
	return -1;
}

/*
 *	DICTIONARY MANAGEMENT
 */
static ONEWjisyo_touroku(touroku,hinshi,eyomi,ekanji)
	Uchar *eyomi,*ekanji;
{	wchar yomi[256],kanji[2048];
	char dicname[1024];
	int rcode;

	ToServerString(eyomi,yomi);
	ToServerString(ekanji,kanji);

	if( touroku ){
		int udic;

		strcpy(dicname,"USERDIC1");
		if( (udic = change_udic(dicname)) < 0 ){
			putmsg(1,"FAILED: cannot open user dictionary [%s]",
				dicname);
			return -1;
		}
		if( (rcode = jd_wreg(kanji,yomi,1<<hinshi)) < 0 ){
			putmsg(1,"(%s) %s",dicname,ONEW_msg_tourokuERROR);
			ONEW_BEEP(2);
		}else	putmsg(1,"(%s) %s",dicname,ONEW_msg_tourokuOK);
	}else{
		dicname[0] = 0;
		get_dicno(dicname,OW_CurrentDic_no);

change_udic(dicname);

		if( 0 <= (rcode = jd_wdel(OW_CurrentSerial,yomi)) )
			putmsg(1,"%s (%s:%d) %s %s",ONEW_msg_massyouOK,
				dicname,OW_CurrentSerial,eyomi,ekanji);
		else{
			putmsg(1,"%s (%s:%d) %s %s",ONEW_msg_cantdelete,
				dicname,OW_CurrentSerial,eyomi,ekanji);
			ONEW_BEEP(2);
		}
	}
	return rcode;
}


static ONEWjisyo_hinshi(hinship)
	Hinshi **hinship;
{
	*hinship = WNN_HINSHI;
	return 1;
}

RegisterKanakan(ONEW_WnnKanakan);
